import numpy as np

class Data:
    def __init__(self, input_order):
        self.input_order = input_order
        self.train = None
        self.valid = None
        self.test = None

    def load_data_from_file(self, fname, cont=False):
        D = open(fname, 'r')
        Data = []
        print fname
        i = 'lol'
        while (i != ''):
            i = D.readline()
            if i == '':
                break;
            if not cont:
                Data.append([[float(x), 0] if float(x) == 1.0 else [float(x), 1] for x in i.split(',')])
            else:
                Data.append([[float(x)] for x in i.split(',')])
        return np.array(Data)


    def load_and_process_train_data(self, fname):
        D = self.load_data_from_file(fname)
        P1 = D[:, self.input_order, :]
        P2 = np.reshape(P1, (len(P1), P1.shape[1]*P1.shape[2]))
        self.train = P2.T

    def load_and_process_train_data_mem(self, fname, cont=False):
        D = self.load_data_from_file(fname, cont)
        self.train = D

    def load_and_process_valid_data(self, fname, cont=False):
        D = self.load_data_from_file(fname, cont)
	self.valid = D

    def load_and_process_test_data(self, fname, cont=False):
        D = self.load_data_from_file(fname, cont)
      #  P1 = D[:, self.input_order, :]
      #  P2 = np.reshape(P1, (len(P1), P1.shape[1]*P1.shape[2]))
        self.test = D# P2.T
        



#loading a graph from a file

from tachyon.nodes import *
import random
import numpy as np
r = lambda x: 0.1 + random.random()*0.1
def generate_children(nodes, curr_node, scope, bf, depth=0, max_depth=10, ctype='c'):
 
    curr_id = int(curr_node.id) + 1
    if isinstance(curr_node, PrdNode):
        num_children = random.randint(bf[0][0], bf[0][1]) if depth > max_depth - 2 else len(scope)
        scopes = [[] for x in range(num_children)]
        weights = np.array([1.0]*num_children)/num_children
        for s in scope:
            slot = np.random.choice(num_children, 1, p=weights)[0]
            val = weights[slot]/4
            weights += (weights[slot] - val)/(len(weights)-1)
            weights[slot] = val
            scopes[slot].append(s)
        scopes = filter(lambda x: x != [], scopes)
	if depth > max_depth - 2:
		scopes = map(lambda x: [x], scope)
        for c in xrange(len(scopes)):
            if len(scopes[c]) == 1:
                a = random.random()
                child = Leaf(str(curr_id+1), a, 1-a, scopes[c][0], t=ctype)
                curr_node.children.append(str(curr_id + 1))
                child.parents.append(curr_node.id)
                curr_id += 1
                nodes.append(child)
            else:
                child = SumNode(str(curr_id + 1))
                child.parents.append(curr_node.id)
                curr_node.children.append(str(curr_id + 1))
                nodes, curr_id = generate_children(nodes, child, scopes[c], bf, depth+1, max_depth, ctype=ctype)
        nodes.append(curr_node)
        return nodes, curr_id+1
    else:
        num_children = random.randint(bf[1][0], bf[1][1])
        scopes = [scope]*num_children
        #print scopes
        for c in xrange(len(scopes)):
            child = PrdNode(str(curr_id + 1))
            curr_node.children.append(str(curr_id + 1))
            curr_node.weights.append(random.random() - 0.5)
            child.parents.append(curr_node.id)
            nodes, curr_id = generate_children(nodes, child, scopes[c], bf, depth+1, max_depth, ctype=ctype)
        norm_factor = sum(curr_node.weights)
        curr_node.weights = map(lambda x: x/norm_factor, curr_node.weights)
        nodes.append(curr_node)
        return nodes, curr_id+1
        

def split_up_file(fname):
    infile = open(fname, 'r')
    t = 'placeholder'
    lines = []
    while t != '':
        t = infile.readline()
        lines.append(t[:-1])
    n = 0
    for i in range(len(lines)):
        if "EDGES" in lines[i] or "Edges" in lines[i]:
            n = i;
            break;

    nodes = lines[1:n]
    edges = lines[n+1:]
    return nodes, edges

def build_nodes(nodes, random_weights=False, ctype="b"):
    big_dict = {}
    Leaves = []
    Prods = []
    Sums = []
    for l in nodes:
        if 'PRD' in l:
            arr = l.split(',')
            node = PrdNode(arr[0])
            big_dict[arr[0]] = node
            Prods.append(arr[0])
        elif 'SUM' in l:
            arr = l.split(',')
            node = SumNode(arr[0])
            big_dict[arr[0]] = node
            Sums.append(arr[0])
        elif 'LEAVE' in l or 'BINN' in l:
            arr = l.split(',')
            if random_weights:
                arr[3] = r(1)
                arr[4] = r(1)
            node = Leaf(arr[0], arr[4], arr[3], arr[2], t=ctype)
            big_dict[arr[0]] = node
            Leaves.append(arr[0])
    return Leaves, Prods, Sums, big_dict


def add_connections(id_node_dict, edges, random_weights=False):
    for edge in edges:
        a = edge.split(',')
        if a[0] == '' or a[1] == '':
            continue
        id_node_dict[a[0]].children.append(a[1])
        id_node_dict[a[1]].parents.append(a[0])
        if len(a) == 3:
            if random_weights:
                id_node_dict[a[0]].weights.append(r(5))
            else:
                id_node_dict[a[0]].weights.append(a[2])
    return id_node_dict

def add_ranks(id_node_dict, leaf_id):
    currs = set(leaf_id)
    rank = 1
    while len(currs) > 0:
        prev_currs = currs
        new_currs = set()
        for s in list(currs):
            for p in id_node_dict[s].parents:
                new_currs.add(p)
            id_node_dict[s].rank = rank
        currs = new_currs
        rank += 1
    orank = rank
    rank -= 1
    currs = prev_currs
    while len(currs) > 0:
        new_currs = set()
        #print rank
        #print currs
        for s in list(currs):
            for p in id_node_dict[s].children:
                new_currs.add(p)
            id_node_dict[s].TRank = rank
        currs = new_currs
        rank -= 1
    return orank, id_node_dict

def create_layers(id_node_dict, rank):
    node_list = [[] for x in range(rank)]
    for k in id_node_dict.keys():
        n = id_node_dict[k]
        #print n.TRank
        node_list[n.TRank].append(n)
    return node_list[1:]

def make_pos_dict(node_layers):
    new_dict = {}
    for i in range(len(node_layers)):
        node_layers[i].sort(lambda x, y: 1 if isinstance(x, Leaf) else -1)
        for j in range(len(node_layers[i])):
            new_dict[node_layers[i][j].id] = (i, j)
    return new_dict, node_layers

def clean_up_inputs(node_layers):
    input_layers = []
    input_order = []
    leaf_order = []
    for n_lst in node_layers:
        c = 0
        for n in n_lst:
            if isinstance(n, Leaf):
                c += 1
                input_order.append(int(n.inp))
                leaf_order.append(n.id)
        input_layers.append(c)
    return leaf_order, input_layers, input_order

def format_list_of_nodes(lst):
    id_node_dict = {}
    leafs = []
    sums = []
    prds = []
    for n in lst:
        if isinstance(n, SumNode):
            sums.append(n.id)
        elif isinstance(n, PrdNode):
            prds.append(n.id)
        else:
            leafs.append(n.id)
        assert id_node_dict.get(n.id, None) == None
        id_node_dict[n.id] = n
    return leafs, prds, sums, id_node_dict

def build_random_net(bf, inp_size, out_size):
    count = 0
    total_net = []
    for i in range(out_size):
        init_node = SumNode(str(count))# if random.random() < 0.51 else PrdNode('0')
        network, count = generate_children([], init_node, range(inp_size), bf)
        total_net += network
    network = total_net
    leaf_ids, prod_ids, sum_ids, id_node_dict = format_list_of_nodes(network)
    print len(id_node_dict)
    #determine all the ranks for each node
    rank, id_node_dict = add_ranks(id_node_dict, leaf_ids)
    #turn them all into layers
    node_layers = create_layers(id_node_dict, rank)
    # print map(lambda x: len(x), node_layers)
    #create a dict for the position of every node given the ids
    pos_dict, node_layers = make_pos_dict(node_layers)
    for n in node_layers:
        print len(n)
    print len(leaf_ids)
    #getting the ordering right
    leaf_id_order, input_layers, input_orders = clean_up_inputs(node_layers)
    print leaf_id_order

    return pos_dict, id_node_dict, node_layers, leaf_id_order, input_layers, input_orders


def load_file(fname, random_weights=False):
    #get the node and edge strings from a file
    file_nodes, file_edges = split_up_file(fname)
    #get all the different nodes and a dict that matches id to node
    leaf_ids, prod_ids, sum_ids, id_node_dict = build_nodes(file_nodes)
    #add all the edges to the nodes
    id_node_dict = add_connections(id_node_dict, file_edges, random_weights)
    if random_weightsi and False:
        for id in sum_ids:
            summ = sum(id_node_dict[id].weights)
            id_node_dict[id].weights = map(lambda x: x/summ, id_node_dict[id].weights)
    #determine all the ranks for each node
    rank, id_node_dict = add_ranks(id_node_dict, leaf_ids)
    #turn them all into layers
    node_layers = create_layers(id_node_dict, rank)
    # print map(lambda x: len(x), node_layers)
    #create a dict for the position of every node given the ids
    pos_dict, node_layers = make_pos_dict(node_layers)
    #getting the ordering right
    leaf_id_order, input_layers, input_orders = clean_up_inputs(node_layers)

    return pos_dict, id_node_dict, node_layers, leaf_id_order, input_layers, input_orders



#Nodes
import numpy as np
class SumNode:
    def __init__(self, id):
        self.id = id;
        self.children = []
        self.parents = []
        self.weights = []
        self.rank = 0
        self.TRank = 0
        self.alive = 1
	self.flags = []
    def dead_children(self, thresh=0):
        a = []
	flags = [0]*len(self.weights)
	for f in self.flags:
		flags[f] = 1
        weights = np.array(self.weights)
        weights /= np.sum(weights)
        new_children = []
        new_weights = []
        print flags
        for c, nw, w in zip(self.children, flags, self.weights):
            if nw < 0.01:
                new_children.append(c)
                new_weights.append(w)
            else:
                a.append(c)
        self.weights = new_weights
        self.children = new_children
        return a

    def serialize_node(self):
        return self.id + ",SUM"

    def serialize_edges(self):
        return [self.id + "," + y + "," + str(x) for x, y in zip(self.weights, self.children)]

class PrdNode:
    def __init__(self, id):
        self.id = id
        self.children = []
        self.parents = []
        self.rank = 0
        self.TRank = 0
        self.alive = 1

    def dead_children(self, thresh=0):
        return []

    def serialize_node(self):
        return self.id + ",PRD"

    def serialize_edges(self):
        return [self.id + "," + x for x in self.children]

class Leaf:
    def __init__(self, id, a, b, i, t='c'):
        self.t = t
        self.id = id;
        self.inp = i;
        self.children = []
        self.parents = [];
        self.weights = [a, b];
        self.rank = 1;
        self.TRank = 0;
        self.alive = 1

    def dead_children(self, thresh=0):
        return []

    def serialize_node(self):
        return self.id + ",LEAVE," + str(self.inp) + "," + str(self.weights[1]) + "," + str(self.weights[0])
    
    def serialize_edges(self):
        return []





#Model file
from tachyon.load_file import *
from tachyon.nodes import *
import random
import tensorflow as tf
import numpy as np
from tachyon.e_load import *

# def findMaxesTensor(T, S):


class Model:

    def __init__(self, optimizer=tf.train.AdamOptimizer, file_weights=True):
        #filename: file to load the model from
        #optimizer: the SGD optimizer prefered
        #file_weights: A boolean representing whether to use weights from file or made up weights
        #currently doesn't allow you to build custom SPNs

        #Tensorflow Graph Variables        
        self.input_swap = None
        self.summ = tf.placeholder(dtype=tf.string, shape=());
        self.loss_summary = None;
        self.writer = None;
        self.input = None #input tensor
        self.output = None #output tensor
        self.sparse_tensors = [] #all sparse tensors to be used in graph
        self.variables = [] #all variables
        self.layer_computations = [] #all mid computations in graph
        self.norm_value = 1.0
        self.counter_tensors = []
        self.weights = []
        self.num_weights = 0
        self.counting = []
        self.updates = []
        self.sum_of_weights = []
        self.labels = None
        self.norm_weights = []
        self.num = None
        self.step_size = 0.01
        #Nodes loaded from files
        self.out_size = 1
        self.pos_dict, self.id_node_dict, self.node_layers, self.leaf_id_order, self.input_layers, self.input_order = None, None, None, None, None, None
        self.cont = [0, 0, 0]
        #For training and inference
        self.loss = None #loss function, None if uninitialized
        self.optimizer = optimizer
        self.optimizer2 = tf.train.AdamOptimizer
        self.opt_val = None
        self.session = None
        self.initalizer = tf.initialize_all_variables
        self.shuffle = None;
        self.reverse_shuffle = None;
        self.inds = None;
        self.multiclass = False;
        self.cccp_updates = []
        self.means = self.vars = []
        #Do Work

    def set_mv(self, data):
        self.mean = np.mean(data, axis=1)
        self.vars = np.var(data, axis=1)

    def build_model_from_file(self, fname, random_weights, mem=False):
        self.pos_dict, self.id_node_dict, self.node_layers, self.leaf_id_order, self.input_layers, self.input_order = load_file(fname, random_weights)

    def build_random_model(self, bfactor, input_length, output_size=1, ctype='b', multiclass=False):
        self.pos_dict, self.id_node_dict, self.node_layers, self.leaf_id_order, self.input_layers, self.input_order, self.shuffle, self.inds = e_build_random_net(bfactor, input_length, output_size, ctype)
        self.multiclass=multiclass
        self.mean = np.array([0]*input_length)
        self.vars = np.array([1]*input_length)

    def build_fast_model(self, bfactor, input_length, ctype='b'):
        self.pos_dict, self.id_node_dict, self.node_layers, self.leaf_id_order, self.input_layers, self.input_order, self.shuffle, self.inds = e_load(bfactor, input_length, ctype)
	self.num_weights = 0;
	for L in self.node_layers[1::2]:
		self.num_weights += len(L)
	self.num_weights += len(self.input_order)


    def compile(self):
        self.build_variables()
        self.build_forward_graph()
        self.start_session()
        self.writer = tf.train.SummaryWriter('logs/minibatch_range3', self.session.graph_def)
        self.get_normal_value()
        self.close_session()

    def build_input_vector(self, leafs):
        ws = []
        inds = []
        for i in range(len(leafs)):
            node = self.id_node_dict[str(leafs[i])]
            a = float(node.weights[0])
            b = float(node.weights[1])
            ws.append(a)
            ws.append(b)
            inds.append(i)
            inds.append(i)
        return tf.Variable(ws, dtype=tf.float64), inds

    def build_reverse_shuffle(self, shuffle):
        reverse_shuffle = []
        for s in shuffle:
            new_s = [0]*len(s)
            for i in range(len(s)):
                new_s[s[i]] = i
            reverse_shuffle.append(new_s)
        return reverse_shuffle

    def unbuild_fast_variables(self):
        weights = self.session.run(self.weights)
        for L in range(1, len(self.node_layers)):
            if isinstance(self.node_layers[L][0], SumNode):
                i = 0
                j = 0
                while i < len(self.node_layers[L]) and isinstance(self.node_layers[L][i], SumNode):
                    for p in range(len(self.node_layers[L][i].weights)):
                        if self.id_node_dict[self.node_layers[L][i].id].weights[p] == weights[L][j+p]:
			    self.id_node_dict[self.node_layers[L][i].id].flags.append(p)
			self.id_node_dict[self.node_layers[L][i].id].weights[p] = weights[L][j+p]
                    j += len(self.node_layers[L][i].weights)
                    i += 1

        for i in range(len(self.leaf_id_order)):
            self.id_node_dict[self.leaf_id_order[i]].weights[0] = weights[0][i*2]
            self.id_node_dict[self.leaf_id_order[i]].weights[1] = weights[0][i*2+1]
    
    def save(self, fname):
        nodes = []
        edges = []
        for v in self.id_node_dict.values():
            if not v.alive: continue
            nodes.append(v.serialize_node())
            edges += v.serialize_edges()
        f = open(fname, 'w')
        f.write("####NODES####")
        for n in nodes:
            f.write("\n")
            f.write(n)
        f.write("\n")
        f.write("####EDGES####")
        for e in edges:
            f.write("\n")
            f.write(e)
        f.write("\n")
        f.close()
        
    def clean_nodes(self, thresh=0.05):
        for v in self.id_node_dict.values():
            if not v.alive: continue
            curr = v.dead_children(thresh)
            while len(curr) > 0:
                node = self.id_node_dict[curr.pop()]
                if not node.alive: continue
                node.alive = 0
                node.parents = []
                curr += node.children
                node.children = []

    def build_fast_variables(self):
        with tf.name_scope("Variables"):
            self.input_swap = tf.Variable(self.input_order)
            weights = []
            if self.node_layers[0][0].t != 'b':
                 self.cont[0] = tf.Variable([self.mean[x] + (lambda y: random.random()*y-(y*0.5))(np.sqrt(self.vars[x])) for x in self.input_order], dtype=tf.float64)
                 self.cont[1] = tf.Variable([self.vars[x] for x in self.input_order], dtype=tf.float64)
                 self.cont[2] = tf.Variable([1.0]*len(self.leaf_id_order), dtype=tf.float64, trainable=False)
            input_weights, input_inds = self.build_input_vector(self.leaf_id_order)
            weights.append(input_weights)
            for L in range(1, len(self.node_layers)):
                ws = []
                if isinstance(self.node_layers[L][0], SumNode):
                    i = 0;
                    j = 0;
                    while i < len(self.node_layers[L]) and isinstance(self.node_layers[L][i], SumNode):
                        ws += map(lambda x: float(x), self.node_layers[L][i].weights);

                        j += len(self.node_layers[L][i].weights)
                        i += 1;
                weights.append(tf.Variable(ws, dtype=tf.float64))
            self.weights = weights;
            self.reverse_shuffle = self.build_reverse_shuffle(self.shuffle)
            self.inds = map(lambda x: tf.Variable(x), [input_inds]+self.inds)
            self.shuffle = map(lambda x: tf.Variable(x), self.shuffle)
            self.reverse_shuffle = map(lambda x: tf.Variable(x), self.reverse_shuffle)
        # self.input_layers = map(lambda x: [0] + tf.constant(x), self.input_layers)

    def get_norm_weights(self):
	weights = filter(lambda x: x != None, self.norm_weights);
	return self.session.run(weights);

    def normalize_weights(self):
        weights = filter(lambda x: x != None, self.norm_weights)
        new_weights = self.session.run(weights);
        old_weights = filter(lambda x: x.get_shape()[0] > 0, self.weights)
        for w, n in zip(old_weights, new_weights):
           # print w.get_shape()
           # print n.shape
            z = w.assign(n)
            self.session.run(z)

    def build_fast_forward_pass(self, step=0.003):
        self.check_op = tf.add_check_numerics_ops()
        computations = []
        bob = 1
        if self.node_layers[0][0].t == 'b':
            bob = 2
        print bob
        with tf.name_scope("input"):
            self.input = tf.placeholder(dtype=tf.float64, 
                                        shape=(None, max(self.input_order)+1, 
                                               bob),
                                        name='Input')
        # self.input = tf.placeholder(dtype=tf.float64, 
        #                                  shape=(len(self.input_order)*2), name='Input')
        #the input to be appended to each layer
        input_splits = []
        self.conz = tf.placeholder(shape=[1], dtype=tf.float64)
        #compute the input
        weights = []
        with tf.name_scope("projection"):
            n = tf.constant(0.0001, dtype=tf.float64)
            for L in range(len(self.weights)):
                if L != 0:
                    drop = tf.round(tf.random_uniform(self.weights[L].get_shape(), self.conz, 1.0, dtype=tf.float64))
                    weights.append(tf.add(tf.nn.relu(tf.sub(self.weights[L]*drop, n)), n))
                else:
                    weights.append(tf.add(tf.nn.relu(tf.sub(self.weights[L], n)), n))

        with tf.name_scope('nomralization'):
            # print (map(lambda x: x.get_shape(), self.inds))
            self.sum_of_weights = [tf.segment_sum(x, y) if x.get_shape()[0] > 0 else None for x, y in zip(weights, self.inds)]
            sum_of_weights = self.sum_of_weights
            self.norm_weights = [tf.div(x, tf.gather(y, z)) if x.get_shape()[0] > 0 else None for x, y, z in zip(weights, self.sum_of_weights, self.inds)]
        # print sums_of_sparses
        
        with tf.name_scope('LEAFS_' + str(len(self.input_order))):
            print len(self.input_order)
            input_gather = tf.reshape(tf.transpose(tf.gather(tf.transpose(self.input, (1, 0, 2)), self.input_swap), (1, 0, 2)), shape=(-1, len(self.input_order)*bob))
            self.counting.append(input_gather)
            # input_gather = tf.Print(input_gather, [input_gather], first_n=25, summarize=25)
            if self.node_layers[0][0].t == 'b':
               input_computation_w = tf.mul(input_gather, weights[0])
               input_computation_s = tf.transpose(tf.segment_sum(tf.transpose(input_computation_w), self.inds[0]))
               input_computation_n = tf.log(tf.div(input_computation_s, sum_of_weights[0]))
               computations.append(input_computation_n)
            else:
               pi = tf.constant(np.pi, tf.float64)
               mus = self.cont[0]
               sigs = tf.nn.relu(self.cont[1] - 0.01) + 0.01
               input_computation_g = tf.div(tf.exp(tf.neg(tf.div(tf.square(input_gather - mus), 2*tf.mul(sigs, sigs)))), tf.sqrt(2*pi)*sigs) + 0.000001
               input_computation_n = tf.log(input_computation_g)
               computations.append(input_computation_n)

        
        #split the input computation and figure out which one goes in each layer
            j = 0
            for i in range(len(self.input_layers)):
                a = tf.constant(j)
                b = self.input_layers[i]
                input_splits.append(tf.slice(input_computation_n, [0, a], [-1, b]))
                # input_splits.append(tf.slice(input_computation_n, [a], [b]))
                j += b;
                # print "size"
                # print size

        current_computation = input_splits[0]

        for i in range(len(self.node_layers[1:])):
            L = i+1 #the layer number
            # print current_computation.get_shape()
            # print self.shuffle[i]
            # print self.inds[L].get_shape()
            # print "yoo"
            
            if self.weights[L].get_shape()[0] == 0: #product
                with tf.name_scope("PRD" + str(self.inds[L].get_shape()[0])):
                    current_computation = tf.transpose(tf.segment_sum(tf.transpose(current_computation), self.inds[L]))
               
            else:
                with tf.name_scope("SUM" + str(self.inds[L].get_shape()[0])):
                    self.counting.append(current_computation)
                    maxes = tf.transpose(tf.segment_max(tf.transpose(current_computation), self.inds[L]))
                    back_maxes = tf.transpose(tf.gather(tf.transpose(maxes), self.inds[L]))
                    current_computation = tf.sub(current_computation, back_maxes)
                    current_computation = tf.exp(current_computation)
                    current_computation = tf.mul(current_computation, weights[L])
                    current_computation = tf.transpose(tf.segment_sum(tf.transpose(current_computation), self.inds[L]))
                    current_computation = tf.div(current_computation, sum_of_weights[L])
                    current_computation = tf.add(tf.log(current_computation), maxes)
                    current_computation = tf.concat(1, [current_computation, input_splits[L]])
                    current_computation = tf.transpose(tf.gather(tf.transpose(current_computation), self.shuffle[L]))
               
            computations.append(current_computation)
        with tf.name_scope('root_node'):
            self.output = current_computation
        with tf.name_scope('loss'):
            if self.multiclass:
                self.labels = tf.placeholder(shape=(None, len(self.node_layers[-1])), dtype=tf.float64)
                self.loss = -tf.reduce_mean(tf.mul(self.output, 0.1*(self.labels-1)+self.labels))
            else:
                self.loss = -tf.reduce_mean(self.output)
            self.loss_summary = tf.scalar_summary(self.summ, self.loss)
        self.opt_val = self.optimizer(0.001).minimize(self.loss)
#        self.opt_val2 = self.optimizer(0.001).minimize(self.loss)
        self.computations = computations

    def cccp(self):
        with tf.name_scope("CCCP"):
            updates =  []
            splits = []
            self.num = tf.placeholder(shape=(None, len(self.node_layers[-1])), dtype=tf.float64)
            curr = self.num
            for i in reversed(range(len(self.node_layers[1:]))):
                L = i+1
                if self.weights[L].get_shape()[0] == 0: #product node
                    curr = tf.transpose(tf.gather(tf.transpose(curr), self.inds[L], name="myprodgather"))
                    factored_comp = tf.transpose(tf.gather(tf.transpose(self.computations[L]), self.inds[L]))
                    curr = curr + factored_comp - self.computations[i][:, :curr.get_shape()[1]]
                  #  print "prd"
                else: #sum node
                  #  print "sum"
                    curr = tf.transpose(tf.gather(tf.transpose(curr), self.reverse_shuffle[L]))
                    if (self.input_layers[L] > 0):
                        curr, split = curr[:, :-self.input_layers[L]], curr[:, -self.input_layers[L]:]
                        splits = [split] + splits;
                    curr = tf.transpose(tf.gather(tf.transpose(curr), self.inds[L], name="mysumgather"))
                    update = self.weights[L]*tf.exp(curr - self.computations[-1] + self.computations[i])
                    curr = curr + tf.log(self.weights[L])
                    updates.append(tf.reduce_sum(update, reduction_indices=0))
            inputs = tf.concat(1, [curr] + splits, name="lolface");
            if self.node_layers[0][0].t == 'b':
                gathered = tf.transpose(tf.gather(tf.transpose(inputs), self.inds[0]))
                update = self.weights[0]*tf.exp(gathered - self.computations[-1])*self.counting[0]+0.000001
                updates.append(tf.reduce_sum(update, reduction_indices=0))
            self.cccp_updates = updates

    def apply_cccp(self, feed_dict, compute_size=100):
        data = feed_dict[self.input]
        updates = []
        for i in range(1+(len(data)-1)//compute_size):
            feed_dict[self.input] = data[i*compute_size:min((i+1)*compute_size, len(data))]
            feed_dict[self.num] = feed_dict[self.num][:-i*compute_size+min((i+1)*compute_size, len(data))]
            update = self.session.run(self.cccp_updates, feed_dict=feed_dict)
            if len(updates) == 0:
                updates = update
            else:
                for i in range(len(update)):
                    updates[i] += update[i]
        # print updates
        a=0
        weights = filter(lambda x: x.get_shape()[0] > 0, self.weights)
        updates = list(reversed(updates))
        for i in range(len(weights)):
            if self.node_layers[0][0].t == 'c' and i == 0:
                for j in range(3):
                    z = tf.assign(self.cont[-j-1], updates[i+j])
                    self.session.run(z)
                a = 2
                continue
            u, w = updates[i+a], weights[i]
            z = tf.assign(w, u)
            self.session.run(z)
        self.normalize_weights()       

    def countss(self):
        print [x.get_shape() for x in self.weights]
        weights = filter(lambda x: x.get_shape()[0] != 0, self.weights)
        with tf.name_scope("Counting"):
            maxed_out = []
            val = tf.constant(0.51, dtype=tf.float64)
            t = lambda x: tf.transpose(x)
            for c in range(len(self.counting)):
                if c == 0 and self.node_layers[0][0].t == 'c':
                    counts = self.counting[c]*0+1
                else:
                    qq = tf.random_uniform(weights[c].get_shape(), 0.9, 1.1, dtype=tf.float64)
                    maxes = tf.mul(tf.transpose(tf.segment_max(tf.transpose(self.counting[c]), self.inds[c*2])), val)
                    back_maxes = tf.transpose(tf.gather(tf.transpose(maxes), self.inds[c*2]))
                    counts = tf.nn.relu(tf.round(tf.div(back_maxes, self.counting[c])))
                maxed_out.append(counts)
            updates = []
            splits = []
            self.num2 = tf.placeholder(shape=(None, len(self.node_layers[-1])), dtype=tf.float64)
            curr = self.num2
            for i in reversed(range(len(self.node_layers[1:]))):
                L = i+1
                if self.weights[L].get_shape()[0] == 0: #product node
                    curr = tf.transpose(tf.gather(tf.transpose(curr), self.inds[L], name="myprodgather"))
                  #  print "prd"
                else: #sum node
                  #  print "sum"
                    curr = tf.transpose(tf.gather(tf.transpose(curr), self.reverse_shuffle[L]))
                    if (self.input_layers[L] > 0):
                        curr, split = curr[:, :-self.input_layers[L]], curr[:, -self.input_layers[L]:]
                        splits = [split] + splits;
                    curr = tf.transpose(tf.gather(tf.transpose(curr), self.inds[L], name="mysumgather"))
                    curr = tf.mul(curr, maxed_out[L//2])
                    updates.append(tf.reduce_sum(curr, reduction_indices=0))
            inputs = tf.concat(1, [curr] + splits, name="lolface");
            if self.node_layers[0][0].t == 'b':
                gathered = tf.transpose(tf.gather(tf.transpose(inputs), self.inds[0]))
                updates.append(tf.reduce_sum(tf.mul(gathered, self.counting[0]), reduction_indices=0))
            else:
                important_inputs = tf.reduce_sum(inputs*self.counting[0], reduction_indices=0)
                total = self.cont[2]+tf.reduce_sum(inputs, reduction_indices=0)
                new_mu = (important_inputs + self.cont[0]*self.cont[2])/total
                mu_diff = new_mu - self.cont[0]
                other_diff = self.counting[0] - self.cont[0]
                new_sig = tf.sqrt((self.cont[1]*self.cont[1]*self.cont[2] + tf.reduce_sum(other_diff*other_diff*inputs, reduction_indices=0))/total - mu_diff*mu_diff)
                new_sums = total
                
                updates.append(new_mu)
                updates.append(new_sig)
                updates.append(new_sums)

        self.updates = updates;

    def apply_count(self, feed_dict, compute_size=100, c=1):
        data = feed_dict[self.input]
        updates = []
        for i in range(1+(len(data)-1)//compute_size):
            feed_dict[self.input] = data[i*compute_size:min((i+1)*compute_size, len(data))]
            feed_dict[self.num2] = feed_dict[self.num2][:-i*compute_size+min((i+1)*compute_size, len(data))]
            update = self.session.run(self.updates, feed_dict=feed_dict)
            if len(updates) == 0:
                updates = update
            else:
                for i in range(len(update)):
                    updates[i] += update[i]
        # print updates
        a=0
        weights = filter(lambda x: x.get_shape()[0] > 0, self.weights)
        print [np.sum(x) for x in weights]
        updates = list(reversed(updates))
        for i in range(len(weights)):
            if self.node_layers[0][0].t == 'c' and i == 0:
                for j in range(3):
                    z = tf.assign(self.cont[-j-1], updates[i+j])
                    self.session.run(z)
                a = 2
                continue
            u, w = updates[i+a], weights[i]
            z = tf.assign_add(w, u)
            self.session.run(z)      
 # print updates

    def fast_compile(self, step, tensorboard_dir="logs/"):
        self.build_fast_variables()
        self.build_fast_forward_pass(step)
        self.countss()
        self.cccp()
        self.out_size = len(self.node_layers[-1])
        self.start_session()
        if tensorboard_dir:
            self.writer = tf.train.SummaryWriter(tensorboard_dir, self.session.graph)
        self.close_session()
    def start_session(self):
        assert self.session == None
	config = tf.ConfigProto()
	config.gpu_options.allocator_type = 'BFC'
	config.gpu_options.allow_growth=True
        self.session = tf.Session(config=config)
        self.session.run(self.initalizer())
        return self.session

    def close_session(self):
        assert self.session != None

        self.session.close()
        self.session = None
        return None

    def build_input_matrix(self, leafs):
        inds = []
        ws = []
        s = []
        for i in range(len(leafs)):
            node = self.id_node_dict[str(leafs[i])]
            a = float(node.weights[0])
            b = float(node.weights[1])
            inds.append([i, i*2])
            inds.append([i, i*2+1])
            ws.append(a)
            ws.append(b)
        s = [len(leafs), len(leafs)*2]
        # print inds
        return tf.Variable(ws, dtype=tf.float64), tf.constant(s, dtype=tf.int64), tf.constant(inds, dtype=tf.int64)
        



import tensorflow as tf
import numpy as np
from tachyon.model import *
from tachyon.data import *

class SPN:
	def __init__(self):
		self.model = None
		self.data = self.data = Data([])
		self.loss = []
		self.val_loss = []
		self.test_loss = None
		self.session_on = False
		self.input_order = []
		self.classify = False
		self.continuous = False
		self.out = 1

	def reshape(self, data):
		n_data = data
		a = 2
		if self.continuous:
			a = 1
		n_data = n_data[:, self.input_order, :a]
		n_data = np.reshape(n_data, (len(n_data), n_data.shape[1]*n_data.shape[2]))
		return n_data

	def make_model_from_file(self, fname, random_weights=False, step=0.003):
		self.model = Model()
		self.model.build_model_from_file(fname, random_weights)
		self.model.compile(step)
		
		self.input_order = self.model.input_order

	def make_random_model(self, bfactor, input_size, output=1, step=0.003, cont=True, classify=False, data=[]):
		self.classify = classify
		self.continuous = cont
		self.model = Model()
		if len(data) > 0:
			self.model.set_mv(data)
		self.out = output
		ctype = 'b'
		if cont:
			ctype = 'c'
		self.model.build_random_model(bfactor, input_size, output, ctype=ctype, multiclass=classify)
		self.model.fast_compile(step)
		self.input_order = self.model.input_order

	def make_fast_model_from_file(self, fname, random_weights=False, step=0.003, cont=True, classify=False):
		self.model = Model()
		self.classify = classify
		self.continuous = cont
		ctype = 'b'
		if cont:
			ctype = 'c'
		self.model.build_fast_model(fname, random_weights, ctype=ctype)
		self.model.fast_compile(step)
		self.input_order = self.model.input_order

	def add_data(self, filename, dataset='train', mem=False, cont=False):
		if dataset == 'train':
			self.data.load_and_process_train_data_mem(filename, cont=cont)
		elif dataset == 'valid':
			self.data.load_and_process_valid_data(filename, cont=cont)
		elif dataset == 'test':
			self.data.load_and_process_test_data(filename, cont=cont)

	def start_session(self):
		self.model.start_session()
		self.session_on = True

	def close_session(self):
		self.model.close_session()
		self.session_on = False

	def predict(self, data, minibatch_size=1000):
		ms = minibatch_size
		tot_loss = 0
		a = 0
		b = 0
		runs = 1+(len(data)-1)//ms
		for i in range(runs):
			print i+1, "/", runs
			b = min(len(data), a + ms)
			n_data = data[a:b, :, :]#
			if self.classify:
                                    #print np.argmax(labels[a:b], axis=1
				feed_dict = {self.model.input: n_data, self.model.labels: labels[a:b], self.model.num: labels[a:b]}
			else:
				feed_dict = {self.model.input: n_data, self.model.num: [[1.0]*self.out]*(b-a)}
			if (a == b):
				break
			if summ == "":	
				loss = self.model.session.run([self.model.loss], feed_dict=feed_dict)
			else:
				feed_dict[self.model.summ] = summ
				summary, loss = self.model.session.run([self.model.loss_summary, self.model.loss], feed_dict=feed_dict)
				self.model.writer.add_summary(summary, epoch*len(data) + b)
			tot_loss += (b-a)*np.sum(loss)
			a += ms
		tot_loss /= float(len(data))
		return tot_lossmodel.session.run(self.model.output, feed_dict=feed_dict)
		return output

	def evaluate(self, data, labels=None,summ="", minibatch_size=1000, epoch=0):
		ms = minibatch_size
		tot_loss = 0
		a = 0
		b = 0
		runs = 1+(len(data)-1)//ms
		for i in range(runs):
			print i+1, "/", runs
			b = min(len(data), a + ms)
			n_data = data[a:b, :, :]#
			if self.classify:
                                    #print np.argmax(labels[a:b], axis=1
				feed_dict = {self.model.conz: [0.9], self.model.input: n_data, self.model.labels: labels[a:b], self.model.num: labels[a:b]}
			else:
				feed_dict = {self.model.conz: [0.9], self.model.input: n_data, self.model.num: [[1.0]*self.out]*(b-a)}
			if (a == b):
				break
			if summ == "":	
				loss = self.model.session.run([self.model.loss], feed_dict=feed_dict)
			else:
				feed_dict[self.model.summ] = summ
				summary, loss = self.model.session.run([self.model.loss_summary, self.model.loss], feed_dict=feed_dict)
				self.model.writer.add_summary(summary, epoch*len(data) + b)
			tot_loss += (b-a)*np.sum(loss)
			a += ms
		tot_loss /= float(len(data))
		return tot_loss

	def test(self, inp):
		feed_dict = {self.model.input: inp}
		vals = self.model.session.run(self.model.computations, feed_dict=feed_dict)
		print vals
		return vals;

	def train(self, epochs, data=[], labels=[], minibatch_size=512, valid_data=[], gd=True, compute_size=1000, count=False, cccp=False, patience=100, summ=True):
		if data == []:
			data = self.data.train
			print data.shape
		if self.classify:
			assert labels != []
		# if (valid):
		# 	val_loss, val_sum = self.evaluate(self.data.valid.T, 'valid_loss')
		# 	self.model.writer.add_summary(val_sum, 0)
		# 	print val_loss
		# if (test):
		# 	test_loss, test_sum = self.evaluate(self.data.test.T, 'test_loss')
		# 	self.model.writer.add_summary(test_sum, 0)
		a = 0
		valid = len(valid_data) > 0 #validation or not!
		bad = 0
		b = 0
		s = 0
		val_sum = ""
		if summ:
			val_sum = "valid"
		if valid:
			prev_valid = self.evaluate(valid_data, summ=val_sum, minibatch_size=compute_size, epoch=0)
		history = dict(train_loss=[], valid_loss=[])
		for e in xrange(epochs):
			a = 0
			b = 0
			print 'Epoch ' + str(e)
			tot_loss = 0
			ms = minibatch_size
			for i in range(1+(len(data)-1)//ms):
				print i+1, "/", 1+(len(data)-1)//ms
				b = min(len(data), a + ms)
				n_data = data[a:b, :, :]
				if self.classify:
					feed_dict = {self.model.conz: [0.745], self.model.input: n_data, self.model.labels: labels[a:b], self.model.num: labels[a:b]}
				else:
					feed_dict = {self.model.conz: [0.745], self.model.input: n_data, self.model.num: [[0.0]*self.out]*(b-a), self.model.num2: [[1.0]*self.out]*(b-a)}
				if (a == b):
					break
				if cccp:
					self.model.apply_cccp(feed_dict, compute_size=compute_size)
				if count:
					self.model.apply_count(feed_dict, compute_size=compute_size)
				if gd:
					_ = self.model.session.run([self.model.check_op, self.model.opt_val], feed_dict = feed_dict)
				if not summ or True:
					loss = 0#self.model.session.run([self.model.loss], feed_dict=feed_dict)
				else:
					feed_dict[self.model.summ] = "batch_loss"
					summary, loss = self.model.session.run([self.model.loss_summary, self.model.loss], feed_dict=feed_dict)
					self.model.writer.add_summary(summary, s)
					s += 1
				tot_loss += (b-a)*np.mean(loss)
				a += ms
			tot_loss /= float(len(data))
			print tot_loss
			history["train_loss"].append(tot_loss)
			if valid:
				valid_loss = self.evaluate(valid_data, summ="valid", minibatch_size=compute_size ,epoch=e+1)
				history["valid_loss"] = valid_loss
				print valid_loss
				if valid_loss > prev_valid:
					bad += 1
					if bad == patience:
						break;
				else:
					bad = 0
					prev_valid = valid_loss

			np.random.shuffle(data)
			

	def get_size(self):
		return len(self.model.id_node_dict) + len(self.model.input_order)
	
	def get_weights(self):
		return self.model.num_weights



#!/usr/bin/env python

from distutils.core import setup

setup(name='Tachyon',
      version='1.0.1',
      description='Fast Sum Product Network Library',
      author='Agastya Kalra',
      author_email='a6kalra@uwaterloo.ca',
      url='https://github.com/KalraA/Tachyon',
      packages=['tachyon'],
      py_modules=['numpy', 'tensorflow']
     )



from tachyon.load_file import *
import copy


def e_add_ranks(id_node_dict, leaf_id):
    currs = set(leaf_id)
    rank = 1
    while len(currs) > 0:
        prev_currs = currs
        new_currs = set()
        for s in list(currs):
            for p in id_node_dict[s].parents:
                new_currs.add(p)
            id_node_dict[s].rank = rank
        currs = new_currs
        rank += 1
    orank = rank
    rank -= 1
    top = None;
    currs = prev_currs
    top = list(currs);
    while len(currs) > 0:
        new_currs = set()
        for s in list(currs):
            for p in id_node_dict[s].children:
                new_currs.add(p)
            id_node_dict[s].TRank = rank
        currs = new_currs
        rank -= 1
    return top, orank, id_node_dict


def e_create_layers(id_node_dict, rank, curr):
    node_list = [[] for x in range(rank)]
    for k in id_node_dict.keys():
        n = id_node_dict[k]
        # print n.TRank
        node_list[n.TRank].append(n)
    node_proper = [[] for x in range(rank)]
    inds_proper = [[] for x in range(rank)]
    inds = range(len(curr))
    curr = [id_node_dict[c] for c in curr]
    while(curr and inds):
        next_curr = []
        next_inds = []
        done = set()
        i = 0;
        for n, ind in zip(curr, inds):
            node_proper[n.TRank].append(n)
            inds_proper[n.TRank].append(ind)
            for c in n.children:
                if c not in done:
                    done.add(c)
                    next_curr.append(id_node_dict[c])
                    next_inds.append(i)
            if (n.children):
            	i += 1;
        curr = next_curr#list(set(next_curr));
        inds = next_inds;

    return node_proper[1:], inds_proper[1:]

def e_finish_layers(layers):
	id_map = lambda z: {y.id: i for i, y in enumerate(z)}
	new_layers = [sorted(copy.deepcopy(lst), (lambda x, y: -1 if isinstance(x, SumNode) and isinstance(y, Leaf) else 1)) for lst in layers]
	layer_ids = [id_map(x) for x in new_layers];
	# print reduce(lambda r, f: r+f.children, new_layers[4], [])
	switches = [[layer_ids[x[0]].get(y.id) for y in x[1]] for x in zip(range(len(layers)), layers)]
	return new_layers, switches

def e_make_pos_dict(node_layers):
    new_dict = {}
    for i in range(len(node_layers)):
        for j in range(len(node_layers[i])):
            new_dict[node_layers[i][j].id] = (i, j)
    return new_dict, node_layers

def e_load(fname, random_weights=False, ctype="b"):
    #get the node and edge strings from a file
    file_nodes, file_edges = split_up_file(fname)
    #get all the different nodes and a dict that matches id to node
    leaf_ids, prod_ids, sum_ids, id_node_dict = build_nodes(file_nodes, random_weights, ctype)
    #add all the edges to the nodes
    id_node_dict = add_connections(id_node_dict, file_edges, random_weights)
    if random_weights:
        for id in sum_ids:
            summ = sum(id_node_dict[id].weights)
            id_node_dict[id].weights = map(lambda x: x/summ, id_node_dict[id].weights)
    #determine all the ranks for each node
    
    top ,rank, id_node_dict = e_add_ranks(id_node_dict, leaf_ids)
    #turn them all into layers
    shuffle_layers, inds = e_create_layers(id_node_dict, rank, top)
    node_layers, shuffle_layers = e_finish_layers(shuffle_layers)
    # print map(lambda x: len(x), node_layers)
    #create a dict for the position of every node given the ids
    pos_dict, node_layers = e_make_pos_dict(node_layers)
    #getting the ordering right
    print "print ur mom"
    leaf_id_order, input_layers, input_orders = clean_up_inputs(node_layers)

    return pos_dict, id_node_dict, node_layers, leaf_id_order, input_layers, input_orders, shuffle_layers, inds

def e_build_random_net(bf, inp_size, out, ctype='b'):
    count = 0
    total_net = []
#    for i in range(out):
    init_node = SumNode(str(count + 1))
    network, count = generate_children([], init_node, range(inp_size), bf, ctype=ctype)
    extras = []
    for i in range(1, out):
        node = SumNode(str(count + i))
        node.children = copy.copy(init_node.children)
        node.weights = [x + random.random()*0.3 - 0.15 for x in copy.copy(init_node.weights)]
        extras.append(node)
        network.append(node)
    leaf_ids, prod_ids, sum_ids, id_node_dict = format_list_of_nodes(network)
    for e in extras:
        for c in e.children:
             id_node_dict[c].parents.append(e.id)
    #print len(id_node_dict)
    #determine all the ranks for each node
    rank, id_node_dict = add_ranks(id_node_dict, leaf_ids)
    #turn them all into layers
    top ,rank, id_node_dict = e_add_ranks(id_node_dict, leaf_ids)
    #turn them all into layers
    shuffle_layers, inds = e_create_layers(id_node_dict, rank, top)
    node_layers, shuffle_layers = e_finish_layers(shuffle_layers)
    # print map(lambda x: len(x), node_layers)
    #create a dict for the position of every node given the ids
    pos_dict, node_layers = e_make_pos_dict(node_layers)
    # print network[-1].children
    # print network[-1].rank
    # print network[-1].TRank
    # print network[-1].id
    # print init_node.children
    # print node_layers[-1]
    shuffle_layers[-2] *= out
    temp = []
    for i in xrange(out):
        temp += [x + i for x in inds[-2]]
    inds[-2] = temp

#    print inds 
#    print shuffle_layers
    #getting the ordering right
    leaf_id_order, input_layers, input_orders = clean_up_inputs(node_layers)
    print "generated a network of size", len(id_node_dict), "with", len(node_layers) ,"layers."
    return pos_dict, id_node_dict, node_layers, leaf_id_order, input_layers, input_orders, shuffle_layers, inds
